@@ -4,26 +4,38 @@ import uuid
 import json
 import numpy as np
 import copy
-import multiprocessing
 import os
+import multiprocessing
 import multiprocessing.pool
 import pandas as pd
+import psutil
+import gc
+import sys
 
 from chefboost.training import Preprocess
 from chefboost.commons import functions, evaluate
 
 #----------------------------------------
 
-class NoDaemonProcess(multiprocessing.Process):
-    # make 'daemon' attribute always return False
-    def _get_daemon(self):
-        return False
-    def _set_daemon(self, value):
-        pass
-    daemon = property(_get_daemon, _set_daemon)
+global decision_rules
 
+class NoDaemonProcess(multiprocessing.Process):
+	# make 'daemon' attribute always return False
+	def _get_daemon(self):
+		return False
+	def _set_daemon(self, value):
+		pass
+	daemon = property(_get_daemon, _set_daemon)
+
+class NoDaemonContext(type(multiprocessing.get_context())):
+	Process = NoDaemonProcess
+	
 class MyPool(multiprocessing.pool.Pool):
-    Process = NoDaemonProcess
+	
+	def __init__(self, *args, **kwargs):
+		kwargs['context'] = NoDaemonContext()
+		super(MyPool, self).__init__(*args, **kwargs)
+	
 #----------------------------------------
 def calculateEntropy(df, config):
 	
@@ -55,24 +67,53 @@ def calculateEntropy(df, config):
 	return entropy
 
 def findDecision(df, config):
+	#information gain for id3, gain ratio for c4.5, gini for cart, chi square for chaid and std for regression
+	algorithm = config['algorithm']
+	
+	resp_obj = findGains(df, config)
+	gains = list(resp_obj["gains"].values())
+	entropy = resp_obj["entropy"]
+	
+	if algorithm == "ID3":
+		winner_index = gains.index(max(gains))
+		metric_value = entropy
+		metric_name = "Entropy"
+	elif algorithm == "C4.5":
+		winner_index = gains.index(max(gains))
+		metric_value = entropy
+		metric_name = "Entropy"
+	elif algorithm == "CART":
+		winner_index = gains.index(min(gains))
+		metric_value = min(gains)
+		metric_name = "Gini"
+	elif algorithm == "CHAID":
+		winner_index = gains.index(max(gains))
+		metric_value = max(gains)
+		metric_name = "ChiSquared"
+	elif algorithm == "Regression":
+		winner_index = gains.index(max(gains))
+		metric_value = max(gains)
+		metric_name = "Std"
+	
+	winner_name = df.columns[winner_index]
+	
+	return winner_name, df.shape[0], metric_value, metric_name
+	
+def findGains(df, config):
 	
 	algorithm = config['algorithm']
 	decision_classes = df["Decision"].unique()
 	
 	#-----------------------------
 	
-	if algorithm == 'Regression':
-		stdev = df['Decision'].std(ddof=0)
-	
 	entropy = 0
 	
 	if algorithm == "ID3" or algorithm == "C4.5":
 		entropy = calculateEntropy(df, config)
-	#print("entropy: ",entropy)
-
+	
 	columns = df.shape[1]; instances = df.shape[0]
 
-	gains = []; gainratios = []; ginis = []; reducted_stdevs = []; chi_squared_values = []
+	gains = []
 
 	for i in range(0, columns-1):
 		column_name = df.columns[i]
@@ -85,7 +126,11 @@ def findDecision(df, config):
 		
 		classes = df[column_name].value_counts()
 		
-		gain = entropy * 1; splitinfo = 0; gini = 0; weighted_stdev = 0; chi_squared_value = 0
+		splitinfo = 0
+		if algorithm == 'ID3' or algorithm == 'C4.5':
+			gain = entropy * 1
+		else:
+			gain = 0
 		
 		for j in range(0, len(classes)):
 			current_class = classes.keys().tolist()[j]
@@ -99,7 +144,6 @@ def findDecision(df, config):
 			
 			if algorithm == 'ID3' or algorithm == 'C4.5':
 				subset_entropy = calculateEntropy(subdataset, config)
-				#print("entropy for this sub dataset is ", subset_entropy)
 				gain = gain - class_probability * subset_entropy			
 			
 			if algorithm == 'C4.5':
@@ -113,7 +157,7 @@ def findDecision(df, config):
 				for k in range(0, len(decision_list)):
 					subgini = subgini - math.pow((decision_list[k]/subset_instances), 2)
 				
-				gini = gini + (subset_instances / instances) * subgini
+				gain = gain + (subset_instances / instances) * subgini
 			
 			elif algorithm == 'CHAID':
 				num_of_decisions = len(decision_classes)
@@ -125,66 +169,52 @@ def findDecision(df, config):
 					
 					chi_square_of_d = math.sqrt(((num_of_d - expected) * (num_of_d - expected)) / expected)
 					
-					chi_squared_value += chi_square_of_d
+					gain += chi_square_of_d
 				
 			elif algorithm == 'Regression':
 				subset_stdev = subdataset['Decision'].std(ddof=0)
-				weighted_stdev = weighted_stdev + (subset_instances/instances)*subset_stdev
+				gain = gain + (subset_instances/instances)*subset_stdev
 		
 		#iterating over classes for loop end
 		#-------------------------------
-		
-		if algorithm == "ID3":
-			gains.append(gain)
-		
-		elif algorithm == "C4.5":
-		
+        
+		if algorithm == 'Regression':
+			stdev = df['Decision'].std(ddof=0)
+			gain = stdev - gain
+		if algorithm == 'C4.5':
 			if splitinfo == 0:
 				splitinfo = 100 #this can be if data set consists of 2 rows and current column consists of 1 class. still decision can be made (decisions for these 2 rows same). set splitinfo to very large value to make gain ratio very small. in this way, we won't find this column as the most dominant one.
-				
-			gainratio = gain / splitinfo
-			gainratios.append(gainratio)
+			gain = gain / splitinfo
 		
-		elif algorithm == "CART":
-			ginis.append(gini)
+		#----------------------------------
 		
-		elif algorithm == "CHAID":
-			chi_squared_values.append(chi_squared_value)
+		gains.append(gain)
 		
-		elif algorithm == 'Regression':
-			reducted_stdev = stdev - weighted_stdev
-			reducted_stdevs.append(reducted_stdev)
+	#-------------------------------------------------
 	
-	#print(df)
-	if algorithm == "ID3":
-		winner_index = gains.index(max(gains))
-		metric_value = entropy
-		metric_name = "Entropy"
-	elif algorithm == "C4.5":
-		winner_index = gainratios.index(max(gainratios))
-		metric_value = entropy
-		metric_name = "Entropy"
-	elif algorithm == "CART":
-		winner_index = ginis.index(min(ginis))
-		metric_value = min(ginis)
-		metric_name = "Gini"
-	elif algorithm == "CHAID":
-		winner_index = chi_squared_values.index(max(chi_squared_values))
-		metric_value = max(chi_squared_values)
-		metric_name = "ChiSquared"
-	elif algorithm == "Regression":
-		winner_index = reducted_stdevs.index(max(reducted_stdevs))
-		metric_value = max(reducted_stdevs)
-		metric_name = "Std"
-	winner_name = df.columns[winner_index]
-
-	return winner_name, df.shape[0], metric_value, metric_name
+	resp_obj = {}
+	resp_obj["gains"] = {}
+	
+	for idx, feature in enumerate(df.columns[0:-1]): #Decision is always the last column
+		#print(idx, feature)    
+		resp_obj["gains"][feature] = gains[idx]
+	
+	resp_obj["entropy"] = entropy
+	
+	return resp_obj
+	
+def createBranchWrapper(func, args):
+	return func(*args)
 
 def createBranch(config, current_class, subdataset, numericColumn, branch_index
-	, winner_index, root, parents, file, dataset_features, num_of_instances, metric):
+	, winner_name, winner_index, root, parents, file, dataset_features, num_of_instances, metric, tree_id = 0, main_process_id = None):
+	
+	custom_rules = []
 	
 	algorithm = config['algorithm']
 	enableAdaboost = config['enableAdaboost']
+	enableGBM = config['enableGBM']
+	max_depth = config['max_depth']
 	enableParallelism = config['enableParallelism']
 	
 	charForResp = "'"
@@ -192,9 +222,7 @@ def createBranch(config, current_class, subdataset, numericColumn, branch_index
 		charForResp = ""
 	
 	#---------------------------
-	
-	json_file = file.split(".")[0]+".json"
-	
+		
 	tmp_root = root * 1
 	parents_raw = copy.copy(parents)
 	
@@ -204,15 +232,16 @@ def createBranch(config, current_class, subdataset, numericColumn, branch_index
 		compareTo = current_class #current class might be <=x or >x in this case
 	else:
 		compareTo = " == '"+str(current_class)+"'"
-	
-	#print(subdataset)
-	
+		
 	terminateBuilding = False
 	
 	#-----------------------------------------------
 	#can decision be made?
 	
-	if enableAdaboost == True:
+	if enableGBM == True and root >= max_depth: #max depth
+		final_decision = subdataset['Decision'].mean()
+		terminateBuilding = True
+	elif enableAdaboost == True:
 		#final_decision = subdataset['Decision'].value_counts().idxmax()
 		final_decision = functions.sign(subdataset['Decision'].mean()) #get average
 		terminateBuilding = True
@@ -227,6 +256,7 @@ def createBranch(config, current_class, subdataset, numericColumn, branch_index
 	#elif algorithm == 'Regression' and subdataset['Decision'].std(ddof=0)/global_stdev < 0.4: #pruning condition
 		final_decision = subdataset['Decision'].mean() #get average
 		terminateBuilding = True
+	
 	#-----------------------------------------------
 	
 	if enableParallelism == True:
@@ -240,27 +270,26 @@ def createBranch(config, current_class, subdataset, numericColumn, branch_index
 	check_rule = check_condition+" obj["+str(winner_index)+"]"+compareTo+":"
 	
 	leaf_id = str(uuid.uuid1())
-	custom_rule_file = "outputs/rules/"+str(leaf_id)+".txt"
-		
+	
 	if enableParallelism != True:
-		
-		#check_rule += " # feature: "+winner_name+", instances: "+str(num_of_instances)+", "+metric_name+": "+str(round(metric, 4))
-		
 		functions.storeRule(file,(functions.formatRule(root),"",check_rule))
 	else:
-
-		sample_rule = "   {\n"
-		sample_rule += "      \"current_level\": "+str(root)+",\n"
-		sample_rule += "      \"leaf_id\": \""+str(leaf_id)+"\",\n"
-		sample_rule += "      \"parents\": \""+parents+"\",\n"
-		sample_rule += "      \"rule\": \""+check_rule+"\",\n"
-		sample_rule += "      \"feature_idx\": "+str(winner_index)+",\n"
-		sample_rule += "      \"instances\": "+str(num_of_instances)+",\n"
-		sample_rule += "      \"metric\": "+str(metric)+"\n"
-		sample_rule += "   }"
-	
-		functions.createFile(custom_rule_file, "")
-		functions.storeRule(custom_rule_file, sample_rule)
+		sample_rule = {}
+		sample_rule["current_level"] = root
+		sample_rule["leaf_id"] = leaf_id
+		sample_rule["parents"] = parents
+		sample_rule["rule"] = check_rule
+		sample_rule["feature_idx"] = winner_index
+		sample_rule["feature_name"] = winner_name
+		sample_rule["instances"] = num_of_instances
+		sample_rule["metric"] = metric
+		sample_rule["return_statement"] = 0
+		sample_rule["tree_id"] = tree_id
+		
+		#json to string
+		sample_rule = json.dumps(sample_rule)
+		
+		custom_rules.append(sample_rule)
 	
 	#-----------------------------------------------
 	
@@ -275,41 +304,59 @@ def createBranch(config, current_class, subdataset, numericColumn, branch_index
 			#serial
 			functions.storeRule(file,(functions.formatRule(root+1),decision_rule))
 		else:
-			#parallel
-			sample_rule = "   , {\n"
-			sample_rule += "      \"current_level\": "+str(root+1)+",\n"
-			sample_rule += "      \"leaf_id\": \""+str(leaf_id)+"\",\n"
-			sample_rule += "      \"parents\": \""+parents+"\",\n"
-			sample_rule += "      \"rule\": \""+decision_rule+"\",\n"
-			sample_rule += "      \"feature_idx\": "+str(winner_index)+",\n"
-			sample_rule += "      \"instances\": "+str(num_of_instances)+",\n"
-			sample_rule += "      \"metric\": 0\n"
-
-			sample_rule += "   }"
+			#parallel			
+			sample_rule = {}
+			sample_rule["current_level"] = root+1
+			sample_rule["leaf_id"] = leaf_id
+			sample_rule["parents"] = parents
+			sample_rule["rule"] = decision_rule
+			sample_rule["feature_idx"] = winner_index
+			sample_rule["feature_name"] = winner_name
+			sample_rule["instances"] = num_of_instances
+			sample_rule["metric"] = 0
+			sample_rule["return_statement"] = 1
+			sample_rule["tree_id"] = tree_id
 			
-			functions.storeRule(custom_rule_file, sample_rule)
+			#json to string
+			sample_rule = json.dumps(sample_rule)
+			
+			custom_rules.append(sample_rule)
 	
 	else: #decision is not made, continue to create branch and leafs
 		root = root + 1 #the following rule will be included by this rule. increase root
 		parents = copy.copy(leaf_id)
 		
-		buildDecisionTree(subdataset, root, file, config, dataset_features
-			, root-1, leaf_id, parents)
-					
+		results = buildDecisionTree(subdataset, root, file, config, dataset_features
+			, root-1, leaf_id, parents, tree_id = tree_id, main_process_id = main_process_id)
+		
+		custom_rules = custom_rules + results
+		
 		root = tmp_root * 1
 		parents = copy.copy(parents_raw)
+	
+	gc.collect()
+	
+	return custom_rules
 
-def buildDecisionTree(df, root, file, config, dataset_features, parent_level = 0, leaf_id = 0, parents = 'root', validation_df = None):
-			
+def buildDecisionTree(df, root, file, config, dataset_features, parent_level = 0, leaf_id = 0, parents = 'root', tree_id = 0, validation_df = None, main_process_id = None):
+		
 	models = []
 	
+	decision_rules = []
+	
+	feature_names = df.columns[0:-1]
+	
 	enableParallelism = config['enableParallelism']
 	algorithm = config['algorithm']
 	
 	json_file = file.split(".")[0]+".json"
 	
+	random_forest_enabled = config['enableRandomForest']
+	enableGBM = config['enableGBM']
+	enableAdaboost = config['enableAdaboost']
+	
 	if root == 1:
-		if config['enableRandomForest'] != True and config['enableGBM'] != True and config['enableAdaboost'] != True:
+		if random_forest_enabled != True and enableGBM != True and enableAdaboost != True:
 			raw_df = df.copy()
 	
 	#--------------------------------------
@@ -317,7 +364,7 @@ def buildDecisionTree(df, root, file, config, dataset_features, parent_level = 0
 	df_copy = df.copy()
 	
 	winner_name, num_of_instances, metric, metric_name = findDecision(df, config)
-	
+		
 	#find winner index, this cannot be returned by find decision because columns dropped in previous steps
 	j = 0 
 	for i in dataset_features:
@@ -332,12 +379,13 @@ def buildDecisionTree(df, root, file, config, dataset_features, parent_level = 0
 	#restoration
 	columns = df.shape[1]
 	for i in range(0, columns-1):
-		column_name = df.columns[i]; column_type = df[column_name].dtypes
+		#column_name = df.columns[i]; column_type = df[column_name].dtypes #numeric field already transformed to object. you cannot check it with df itself, you should check df_copy
+		column_name = df_copy.columns[i]; column_type = df_copy[column_name].dtypes
 		if column_type != 'object' and column_name != winner_name:
 			df[column_name] = df_copy[column_name]
 	
 	classes = df[winner_name].value_counts().keys().tolist()
-		
+	#print("classes: ",classes," in ", winner_name)
 	#-----------------------------------------------------
 	
 	num_cores = config["num_cores"]
@@ -355,15 +403,26 @@ def buildDecisionTree(df, root, file, config, dataset_features, parent_level = 0
 		if enableParallelism != True:
 			
 			if i == 0:
-				descriptor = "# Feature: "+winner_name+", Instances: "+str(num_of_instances)+", "+metric_name+": "+str(round(metric, 4))
+				
+				descriptor = {
+					"feature": winner_name,
+					"instances": num_of_instances,
+					#"metric_name": metric_name,
+					"metric_value": round(metric, 4),
+					"depth": parent_level + 1
+				}
+				descriptor = "# "+json.dumps(descriptor)
 				
 				functions.storeRule(file, (functions.formatRule(root), "", descriptor))
 			
-			createBranch(config, current_class, subdataset, numericColumn, branch_index
-				, winner_index, root, parents, file, dataset_features, num_of_instances, metric)
+			results = createBranch(config, current_class, subdataset, numericColumn, branch_index
+				, winner_name, winner_index, root, parents, file, dataset_features, num_of_instances, metric, tree_id = tree_id, main_process_id = main_process_id)
+			
+			decision_rules = decision_rules + results
+			
 		else:
 			input_params.append((config, current_class, subdataset, numericColumn, branch_index
-				, winner_index, root, parents, file, dataset_features, num_of_instances, metric))
+				, winner_name, winner_index, root, parents, file, dataset_features, num_of_instances, metric, tree_id, main_process_id))
 	
 	#---------------------------
 	#add else condition in the decision tree
@@ -380,108 +439,167 @@ def buildDecisionTree(df, root, file, config, dataset_features, parent_level = 0
 			functions.storeRule(file,(functions.formatRule(root+1), else_decision))
 		else: #parallelism
 			leaf_id = str(uuid.uuid1())
-			custom_rule_file = "outputs/rules/"+str(leaf_id)+".txt"
 			
 			check_rule = "else: "+else_decision
 			
-			sample_rule = "   {\n"
-			sample_rule += "      \"current_level\": "+str(root)+",\n"
-			sample_rule += "      \"leaf_id\": \""+str(leaf_id)+"\",\n"
-			sample_rule += "      \"parents\": \""+parents+"\",\n"
-			sample_rule += "      \"rule\": \""+check_rule+"\",\n"
-			sample_rule += "      \"feature_idx\": -1,\n"
-			sample_rule += "      \"instances\": "+str(df.shape[0])+",\n"
-			sample_rule += "      \"metric\": 0\n"
-			sample_rule += "   }"
+			sample_rule = {}
+			sample_rule["current_level"] = root
+			sample_rule["leaf_id"] = leaf_id
+			sample_rule["parents"] = parents
+			sample_rule["rule"] = check_rule
+			sample_rule["feature_idx"] = -1
+			sample_rule["feature_name"] = ""
+			sample_rule["instances"] = df.shape[0]
+			sample_rule["metric"] = 0
+			sample_rule["return_statement"] = 0
+			sample_rule["tree_id"] = tree_id
 			
-			functions.createFile(custom_rule_file, "")
-			functions.storeRule(custom_rule_file, sample_rule)
+			#json to string
+			sample_rule = json.dumps(sample_rule)
+			decision_rules.append(sample_rule)
 			
 	else: #regression
 		else_decision = "return %s" % (subdataset.Decision.mean())
-		
+				
 		if enableParallelism != True:
 			functions.storeRule(file,(functions.formatRule(root), "else:"))
 			functions.storeRule(file,(functions.formatRule(root+1), else_decision))
 		else:
 			leaf_id = str(uuid.uuid1())
-			custom_rule_file = "outputs/rules/"+str(leaf_id)+".txt"
 			
 			check_rule = "else: "+else_decision
 			
-			sample_rule = "   {\n"
-			sample_rule += "      \"current_level\": "+str(root)+",\n"
-			sample_rule += "      \"leaf_id\": \""+str(leaf_id)+"\",\n"
-			sample_rule += "      \"parents\": \""+parents+"\",\n"
-			sample_rule += "      \"rule\": \""+check_rule+"\"\n"
-			sample_rule += "   }"
+			sample_rule = {}
+			sample_rule["current_level"] = root
+			sample_rule["leaf_id"] = leaf_id
+			sample_rule["parents"] = parents
+			sample_rule["rule"] = check_rule
+			sample_rule["tree_id"] = tree_id
+			sample_rule["feature_name"] = ""
+			sample_rule["instances"] = 0
+			sample_rule["metric"] = 0
+			sample_rule["return_statement"] = 1
 			
-			functions.createFile(custom_rule_file, "")
-			functions.storeRule(custom_rule_file, sample_rule)
+			#json to string
+			sample_rule = json.dumps(sample_rule)
+			decision_rules.append(sample_rule)
 	
 	#---------------------------
+		
+	try:
+		main_process = psutil.Process(main_process_id)
+		children = main_process.children(recursive=True)
+		active_processes = len(children) + 1 #plus parent
+		#active_processes = len(children)
+	except:
+		active_processes = 100 #set a large initial value
 	
+	results = []
 	#create branches in parallel
 	if enableParallelism == True:
-		"""
-		#this usage causes trouble for recursive functions
-		with Pool(number_of_cpus) as pool:
-			pool.starmap(createBranch, input_params)
-		"""
 		
-		pool = MyPool(num_cores)
-		results = pool.starmap(createBranch, input_params)
-		pool.close()
-		pool.join()
+		if parent_level == 0 and random_forest_enabled != True:
+		#if main_process_id != None and num_cores >= active_processes + len(classes): #len(classes) branches will be run in parallel #this causes hang and deadlock
+			
+			#--------------------------------
+			"""
+			#causes hang problem if number of input_params is greater than num_cores
+			pool = MyPool(num_cores)
+			branch_results = pool.starmap(createBranch, input_params)
+			
+			for branch_result in branch_results:
+				for leaf_result in branch_result:
+					results.append(leaf_result)
+			
+			pool.close()
+			pool.join()
+			pool.terminate()
+			
+			gc.collect()
+			"""
+			#--------------------------------
+			
+			#workaround for hang problem. set num_cores and active threads same.
+			#len(classes) == len(input_params)
+			#e.g. len(input_params) = 5, num_cores = 2, cycles = 3
+			#we will feed 2 items to pool in for loops instead of 5
+			
+			cycles = int(len(input_params) / num_cores) + 1
+			
+			for i in range(0, cycles):
+				
+				filter_begin = i * num_cores
+				filter_end = i * num_cores + num_cores
+				
+				if filter_end > len(input_params):
+					filter_end = filter_end
+				
+				input_frame = input_params[filter_begin: filter_end]
+				
+				pool = MyPool(num_cores)
+				branch_results = pool.starmap(createBranch, input_frame)
+				
+				pool.close()
+				pool.join()
+				pool.terminate()
+				gc.collect()
+				
+				for branch_result in branch_results:
+					for leaf_result in branch_result:
+						results.append(leaf_result)
+			
+			#--------------------------------
+		
+		else:
+			for input_param in input_params:
+				sub_results = createBranchWrapper(createBranch, input_param)
+				for sub_result in sub_results:
+					results.append(sub_result)
+		
+		#--------------------------------
+		
+		decision_rules = decision_rules + results
+		
+		#--------------------------------
+		
+		if root != 1: #return children results until the root node
+			return decision_rules
 	
 	#---------------------------------------------
 	
-	#calculate accuracy metrics
 	if root == 1:
 		
 		if enableParallelism == True:
-
-			#custom rules are stored in .txt files. merge them all in a json file
 			
-			functions.createFile(json_file, "[\n")
-			
-			custom_rules = []
+			#custom rules are stored in decision_rules. merge them all in a json file first
+						
+			json_rules = "[\n" #initialize
 			
 			file_index = 0
-			for file in os.listdir(os.getcwd()+"/outputs/rules"):
-				if file.endswith(".txt"):
-					custom_rules.append(os.getcwd()+"/outputs/rules/"+file)
-					#print(file) #this file stores a custom rule
-					f = open(os.getcwd()+"/outputs/rules/"+file, "r")
-					custom_rule = f.read()
-					
-					if file_index > 0:
-						custom_rule = ", "+custom_rule
-					
-					functions.storeRule(json_file, custom_rule)
-					f.close()
-					file_index = file_index + 1
-					
-			functions.storeRule(json_file, "]")
+			for custom_rule in decision_rules:
+				
+				json_rules += custom_rule
+				
+				if file_index < len(decision_rules) - 1:
+					json_rules += ", "
+				
+				json_rules += "\n"
+				
+				file_index = file_index + 1				
 			
 			#-----------------------------------
 			
-			#custom rules are already merged in a json file. clear messy custom rules
-			#TO-DO: if random forest trees are handled in parallel, this would be a problem. You cannot know the related tree of a rule. You should store a global tree id in a rule.
-			
-			for file in custom_rules:
-				os.remove(file)
+			json_rules += "]"
+			functions.createFile(json_file, json_rules)
 			
 			#-----------------------------------
+			#reconstruct rules from json to py
 			
-			reconstructRules(json_file)
-
-			#feature importance should be calculated by demand?
-			feature_importance(json_file, dataset_features)
+			reconstructRules(json_file, feature_names)
 			
 			#-----------------------------------
 		
-		#Accuracy calculation
+		#is regular decision tree
 		if config['enableRandomForest'] != True and config['enableGBM'] != True and config['enableAdaboost'] != True:
 		#this is reguler decision tree. find accuracy here.
 			
@@ -490,26 +608,6 @@ def buildDecisionTree(df, root, file, config, dataset_features, parent_level = 0
 			myrules = imp.load_module(moduleName, fp, pathname, description) #rules0
 			models.append(myrules)
 			
-			#--------------------------------
-			#train accuracy
-			
-			#some numerical features transform to nominal. e.g. 85 -> >=80
-			df = raw_df.copy()
-			
-			#instead of for loops, pandas functions perform well
-			df['Prediction'] = df.apply(findPrediction, axis=1)
-			
-			evaluate.evaluate(df)
-			
-			#--------------------------------
-			#validation accuracy
-			
-			if isinstance(validation_df, pd.DataFrame):
-				validation_df['Prediction'] = validation_df.apply(findPrediction, axis=1)
-				evaluate.evaluate(validation_df, task = 'validation')
-			
-			#--------------------------------
-			
 	return models
 			
 def findPrediction(row):
@@ -529,14 +627,27 @@ def findPrediction(row):
 If you set parelellisim True, then branches will be created parallel. Rules are stored in a json file randomly. This program reconstructs built rules in a tree form. In this way, we can build decision trees faster.
 """
 
-def reconstructRules(source):
+def reconstructRules(source, feature_names, tree_id = 0):
 	
 	#print("Reconstructing ",source)
 	
 	file_name = source.split(".json")[0]
 	file_name = file_name+".py"
 	
-	functions.createFile(file_name, "#This rule was reconstructed from "+source+"\n")
+	#-----------------------------------
+	
+	constructor = "def findDecision(obj): #"
+	idx = 0
+	for feature in feature_names:
+		constructor = constructor + "obj["+str(idx)+"]: "+feature
+		
+		if idx < len(feature_names) - 1:
+			constructor = constructor+", "
+		idx = idx + 1
+	
+	functions.createFile(file_name, constructor+"\n")
+	
+	#-----------------------------------
 	
 	with open(source, 'r') as f:
 		rules = json.load(f)
@@ -561,6 +672,10 @@ def reconstructRules(source):
 			rule.append(instance["leaf_id"])
 			rule.append(instance["parents"])
 			rule.append(instance["rule"])
+			rule.append(instance["feature_name"])
+			rule.append(instance["instances"])
+			rule.append(instance["metric"])
+			rule.append(instance["return_statement"])
 			rule_set.append(rule)
 			#print(padleft(instance["rule"], instance["current_level"]))
 
@@ -574,9 +689,14 @@ def reconstructRules(source):
 		
 		leaf_idx = 0
 		for i in range(0 ,df.shape[0]):
+			current_level = int(df[i][0])
 			leaf_id = df[i][1]
 			parent_id = df[i][2]
 			rule = df[i][3]
+			feature_name = df[i][4]
+			instances = int(df[i][5])
+			metric = float(df[i][6])
+			return_statement = int(df[i][7])
 			
 			if parent_id == parent:
 				
@@ -598,6 +718,15 @@ def reconstructRules(source):
 					
 					#print(padleft(rule, level), "(", leaf_idx,")")
 					
+					if leaf_idx == 0 and return_statement == 0:
+						explainer = {}
+						explainer["feature"] = feature_name
+						explainer["instances"] = instances
+						explainer["metric_value"] = round(metric, 4)
+						explainer["depth"] = current_level
+						explainer = "# "+json.dumps(explainer)
+						functions.storeRule(file_name, padleft(explainer, level))
+					
 					functions.storeRule(file_name, padleft(rule, level))
 					
 					level = level + 1; parent = copy.copy(leaf_id)
@@ -614,93 +743,6 @@ def reconstructRules(source):
 			
 	#------------------------------------
 	
-	#print("def findDecision(obj):")
-	functions.storeRule(file_name, "def findDecision(obj):")
 	extractRules(df)
 
 	#------------------------------------
-
-def feature_importance(source, features):
-	
-	with open(source, 'r') as f:
-		rules = json.load(f)
-	
-	rule_set = []
-	for instance in rules:
-		if len(instance) > 0:
-			rule = []
-			if "metric" in list(instance.keys()):
-				rule.append(instance["current_level"])
-				rule.append(instance["leaf_id"])
-				rule.append(instance["parents"])
-				rule.append(instance["rule"])
-				rule.append(instance["feature_idx"])
-				rule.append(instance["instances"])
-				rule.append(instance["metric"])
-				rule_set.append(rule)
-	
-	df = pd.DataFrame(rule_set
-		, columns = ["current_level", "leaf_id", "parents", "rule", "feature_idx", "instances", "metric"])
-
-	feature_importances = []
-	feature_idx = 0
-	for feature in features:
-
-		#print("Feature ", feature)
-		
-		feature_nodes = df[df.feature_idx == feature_idx]
-		feature_nodes = feature_nodes.merge(df, left_on = ["leaf_id"], right_on = ["parents"], how = "left")
-		feature_nodes = feature_nodes[(feature_nodes.feature_idx_y != -1)] #discard else conditions
-
-		#----------------------------------
-
-		pivot = feature_nodes.groupby(by = ["parents_y"])[["parents_x", "metric_x", "instances_x", "metric_y", "instances_y"]].min()
-
-		parents = pivot.parents_x.unique()
-
-		importance = 0
-
-		#print("Feature ", feature, ": ")
-
-		for parent in parents:
-			node_importance = 0
-			child = pivot[pivot.parents_x == parent]
-
-			parent_effect = (child.iloc[0].metric_x * child.iloc[0].instances_x)
-			node_importance = node_importance + parent_effect
-
-			#print(child.iloc[0].metric_x," x ", child.iloc[0].instances_x, end = '')
-
-			for index, instance in child.iterrows():
-				#print(" - ", instance.metric_y, " x ", instance.instances_y, end = '')
-				node_importance = node_importance - (instance.metric_y * instance.instances_y)
-
-			#importance = importance + (parent_effect - child_effect)
-
-			#print(" = ",node_importance)
-			importance = importance + node_importance
-
-		#print("Importance of feature ", feature," is ", importance,". \n")
-
-		feature_importance = []
-		feature_importance.append(feature)
-		feature_importance.append(importance)
-		feature_importances.append(feature_importance)
-
-		feature_idx = feature_idx + 1
-
-	#---------------------------------
-
-	feature_importances = pd.DataFrame(feature_importances, columns = ["feature", "importance"])
-	feature_importances = feature_importances.sort_values(by = ["importance"], ascending = False).reset_index(drop = True)
-	
-	#print(feature_importances)
-	
-	feature_importances.importance = (feature_importances.importance) / feature_importances.importance.sum()
-
-	target = source.split(".")[0]+"_fi.csv"
-
-	feature_importances.to_csv(target, index = False)
-
-	#print(feature_importances)
-	
\ No newline at end of file
